import maya.standalone
import maya.cmds as cmds
import maya.mel as mm
from math import sqrt
import maya.OpenMaya as om
from math import fmod
import sys

def projectRay(cameraName, imageXRes, imageYRes, imageX, imageY, frameNo):
	
    # set time
    cmds.currentTime(frameNo)
	
    worldPt = imageToWorld(cameraName, imageXRes, imageYRes, imageX, imageY)
    camPos = cmds.camera(cameraName, q=True, p=True)
    hitPt = detectCollision(camPos[0], camPos[1], camPos[2], worldPt[0], worldPt[1], worldPt[2])
    #cmds.curve(degree = 1, point = [(camPos[0], camPos[1], camPos[2]), (hitPt[0], hitPt[1], hitPt[2])] )
    return hitPt
    
		
def floatMMatrixToMMatrix_(fm):
    mat = om.MMatrix()
    om.MScriptUtil.createMatrixFromList ([
        fm(0,0),fm(0, 1),fm(0, 2),fm(0, 3),
        fm(1,0),fm(1, 1),fm(1, 2),fm(1, 3),
        fm(2,0),fm(2, 1),fm(2, 2),fm(2, 3),
        fm(3,0),fm(3, 1),fm(3, 2),fm(3, 3)], mat)
    return mat

	
def detectCollision(camX, camY, camZ, woX, woY, woZ):
    
    objects = cmds.ls( type='mesh')
    hitMade = False
    smallestDis = 100000000 
    for i in range(0, len(objects)):
	direction = [woX-camX, woY-camY, woZ-camZ]
	rtn = rayIntersect(objects[i], [camX, camY, camZ], direction)
	if rtn != []:
		d = sqrt(pow(rtn[2]-camZ, 2) + pow(rtn[1]-camY,2) + pow(rtn[0]-camX, 2))
            	if (d < smallestDis):
                	hitMade = True
                	smallestDis = d
                	hitPt = [rtn[0], rtn[1], rtn[2]] 
				
            
    if hitMade:
	return hitPt	
    else:
	return None
        
def detectCollisionBack(camX, camY, camZ, woX, woY, woZ):
    
    objects = cmds.ls( type='mesh')
    hitMade = False
    smallestDis = 100000000 
	
    for i in range(0, len(objects)):
        direction = [woX-camX, woY-camY, woZ-camZ]
	rtn = rayIntersect(objects[i], [camX, camY, camZ], direction)
	if rtn != []:            
            dHitCam = sqrt(pow(rtn[2]-camZ, 2) + pow(rtn[1]-camY,2) + pow(rtn[0]-camX, 2))
            dHitWorld = sqrt(pow(rtn[2]-woZ, 2) + pow(rtn[1]-woY,2) + pow(rtn[0]-woX, 2))
            dCamWorld = sqrt(pow(camZ-woZ, 2) + pow(camY-woY,2) + pow(camX-woX, 2))
            
            if ((abs((dHitCam + dHitWorld) - dCamWorld) < 0.00001) & (dHitCam > 0.00001)  & (dHitWorld > 0.00001)):
                hitMade = True
                hitPt = [rtn[0], rtn[1], rtn[2]]             
            
    if hitMade:
        return hitPt	
    else:
        return None              

        
def imageToWorld(cameraName, imageXRes, imageYRes, imageX, imageY):
    imageX = float(imageX)
    imageY = float(imageY)

    sl = om.MSelectionList()
    sl.add(cameraName)
    dpathCameraShape=om.MDagPath()
    sl.getDagPath(0, dpathCameraShape)
    dpathCameraShape.extendToShape()
    fnc = om.MFnCamera(dpathCameraShape.node())
    
    projMatInv= floatMMatrixToMMatrix_(fnc.projectionMatrix().inverse())   
    camMat = dpathCameraShape.inclusiveMatrix()
    
    proXW = 2.0 * (imageX/imageXRes -0.5)
    proYW = 2.0 * (((imageYRes - imageY)/imageYRes)-0.5)
    projected = om.MPoint(proXW, proYW, -1.0, 1.0)

    world = projected *  projMatInv* camMat
    world = (world[0]/world[3], world[1]/world[3], world[2]/world[3], world[3]/world[3])
    	
    return world[0], world[1], world[2]
    
    
def worldToImage(cameraName, imageXRes, imageYRes, worldX, worldY, worldZ, frameNo):
	
    # set time
    cmds.currentTime(frameNo)

    # check if object is occluded in second frame    
    camPos = cmds.camera(cameraName, q=True, p=True)
    hitPt = detectCollisionBack(worldX, worldY, worldZ, camPos[0], camPos[1], camPos[2])
  
    if (hitPt == None):
        p = om.MPoint(worldX, worldY, worldZ)
        sl = om.MSelectionList()
        sl.add(cameraName)
        dpathCameraShape=om.MDagPath()
        sl.getDagPath(0, dpathCameraShape)
        dpathCameraShape.extendToShape()
        fnc = om.MFnCamera(dpathCameraShape.node())
            
        projMat = floatMMatrixToMMatrix_(fnc.projectionMatrix()) 
        camMatInv = dpathCameraShape.inclusiveMatrix().inverse();          
        projected = p * camMatInv * projMat
            
        imageX = (projected.x/projected.w / 2.0 + 0.5) * imageXRes
        imageY = imageYRes - ((projected.y/projected.w / 2.0 + 0.5) * imageYRes)
        return imageX, imageY
    
    return None

def rayIntersect(mesh, point, direction):

	#posted on cgtalk.com
	om.MGlobal.clearSelectionList()

	om.MGlobal.selectByName(mesh)
	sList = om.MSelectionList()
	#Assign current selection to the selection list object
	om.MGlobal.getActiveSelectionList(sList)

	item = om.MDagPath()
	sList.getDagPath(0, item)
	item.extendToShape()

	fnMesh = om.MFnMesh(item)

	raySource = om.MFloatPoint(point[0], point[1], point[2], 1.0)
	rayDir = om.MFloatVector(direction[0], direction[1], direction[2])
	faceIds = None
	triIds = None
	idsSorted = False
	testBothDirections = False
	worldSpace = om.MSpace.kWorld
	maxParam = 999999
	accelParams = None
	sortHits = True
	hitPoints = om.MFloatPointArray()
	#hitRayParams = om.MScriptUtil().asFloatPtr()
	hitRayParams = om.MFloatArray()
	hitFaces = om.MIntArray()
	hitTris = None
	hitBarys1 = None
	hitBarys2 = None
	tolerance = 0.0001
	hit = fnMesh.allIntersections(raySource, rayDir, faceIds, triIds, idsSorted, worldSpace, maxParam, testBothDirections, accelParams,	sortHits, hitPoints, hitRayParams, hitFaces, hitTris, hitBarys1, hitBarys2, tolerance)

	#clear selection as may cause problem if the function is called	multiple times in succession
	om.MGlobal.clearSelectionList()
		
	# only want closest point
	result = []	
	smallestDis = 9999999999
	for x in range(hitPoints.length()):
		d = sqrt(pow(hitPoints[x][2]-point[2], 2) + pow(hitPoints[x][1]-point[1],2) + pow(hitPoints[x][0]-point[0], 2))
		if (d < smallestDis):
			smallestDis = d
			result = [hitPoints[x][0], hitPoints[x][1], hitPoints[x][2]]
		
			
	return result
    
def getFlow(opDir, seq, cameraName, imageXRes, imageYRes, frame1, frame2):
    
    maya.standalone.initialize(name='python')
    cmds.file(seq, force=True, open=True)
    
    flowF = open(opDir + '/flow_' + seq[:-3] + '_'  + frame1 + '_'  + frame2 + '.txt', 'w')
        
    imageXRes = int(imageXRes)
    imageYRes = int(imageYRes)
    frame1 = int(frame1)
    frame2 = int(frame2)
    
    #looping column order
    for xIt in range (0, imageXRes):
        x = xIt + 0.5
        for yIt in range (0, imageYRes):
            y = yIt + 0.5
            im1W = projectRay(cameraName, imageXRes, imageYRes, x, y, frame1)
            if (im1W != None):
                im2Pt =  worldToImage(cameraName, imageXRes, imageYRes, im1W[0], im1W[1], im1W[2], frame2)
                if (im2Pt != None):
                    if ((im2Pt[0] > 0.5)  & (im2Pt[0] < (imageXRes-0.5)) & (im2Pt[1] > 0.5)  & (im2Pt[1] < (imageYRes-0.5))):
                        flowF.write(repr(im2Pt[0] - x) + ' ' + repr(im2Pt[1] - y) + '\n')
                    else:
                        flowF.write('NaN NaN \n') #no flow
                else:
                    flowF.write('NaN NaN \n') #no flow
            else:
                flowF.write('NaN NaN \n') #no flow

    flowF.close()

# example run
getFlow(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6], sys.argv[7])
# opDir
# sequence name
# camera name
# x res
# y res 
# frame1 
# frame2







